/*
 * Automatically generated by asn1_compiler.  Do not edit
 *
 * ASN.1 parser for x509
 */
#include <linux/asn1_decoder.h>

extern const struct asn1_decoder x509_decoder;

extern int x509_extract_key_data(void *, size_t, unsigned char, const void *, size_t);
extern int x509_extract_name_segment(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_OID(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_issuer(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_not_after(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_not_before(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_params(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_pkey_algo(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_serial(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_signature(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_subject(void *, size_t, unsigned char, const void *, size_t);
extern int x509_note_tbs_certificate(void *, size_t, unsigned char, const void *, size_t);
extern int x509_process_extension(void *, size_t, unsigned char, const void *, size_t);

